import sys, yaml, os, io
from os.path import exists, join, isdir, dirname, isabs, abspath

import logging
log = logging.getLogger('elastico.config')

from .util import string, PY3

from argdeco import ConfigDict
class Undefined:
    pass

from .formatter import ElasticoFormatter

formatter = ElasticoFormatter()

class Config(ConfigDict):
    @classmethod
    def object(cls, value=None, file=None):
        if value is None:
            value = {}

        if isinstance(value, string):
            if not PY3:
                if not isinstance(value, unicode):
                    value = value.decode('utf-8')
            value = yaml.load(io.StringIO(value))

        cfg = cls()
        cfg.update(value)
        cfg.set_filename(file)
        return cfg

    @classmethod
    def create(cls, config_file=None, arguments=None, **kwargs):
        if arguments is None:
            arguments = {}

        if kwargs:
            arguments.update(kwargs)

        if arguments:
            cfg = cls({'arguments': arguments})
        else:
            cfg = cls()

        if config_file:
            cfg.load_from_file(config_file)

        return cfg

    def load_from_file(self, file_name):
        '''argdeco's compiler_factory API'''
        file_name = abspath(file_name)
        self.config_file = file_name
        self.include_file(file_name)
        self.set_filename(file_name)

        self.logging_setup()

    def logging_setup(self):
        logspec = self.get('logging', {})
        if logspec:
            logspec = self.flatten(logspec)
            for k,v in logspec.items():
                if k == 'ROOT':
                    k = None
                logger = logging.getLogger(k)
                logger.setLevel(getattr(logging, v))
                log.info("change loglevel -- logger=%s, level=%s", k, v)

    def refresh(self, **kwargs):
        '''refresh this dictionary by rereading the data from disk'''
        log.info("refresh config -- config_file=%s", self.config_file)
        _arguments = self.get('arguments', {})
        _arguments.update(kwargs)
        self.clear()
        self.load_from_file(self.config_file)
        if _arguments:
            self['arguments'] = _arguments


    def get_filename(self, *configs):
        for config in configs:
            if hasattr(config, '_file'):
                return config
        return self._file

    def set_filename(self, filename):
        if filename is not None:
            self._file = filename
            self._dir = dirname(filename)

    def __getattr__(self, name):
        if name == '_dir':
            self._dir = '.'
            return self._dir
        if name == '_file':
            self._file = '-'
            return self._file
        if name == '_files':
            self._files = set()
            return self._files

        try:
            return self[name]
        except KeyError:
            pass
        raise AttributeError(name)

    def __getitem__(self, name):
        '''first look in self['arguments'] for finding a value.
        '''
#        log.debug("__getitem__: %s", name)

        if name == '.':
            return self

        try:
            arguments = super(Config, self).__getitem__('arguments')
        except KeyError:
            if name == 'arguments': raise
            arguments = {}

        if name == 'arguments':
            result = arguments
        else:
            try:
                result = arguments[name]
            except KeyError:
                result = super(Config, self).__getitem__(name)
        return result

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def update_from_includes(self):
        log.debug("update_from_includes starts: %s", self._file)
        for item in self.get('include', []):
            log.debug("update_from_includes item: %s", item)
            if isinstance(item, string):
                if item.endswith('/'):
                    item = {'directory': item}
                else:
                    item = {'file': item}
            update_name = item.get('update')
            append_name = item.get('append')
            assert not (update_name and append_name), "you can only specify append OR update"
            if append_name:
                action='append'
                name = append_name
            else:
                action='update'
                name = update_name

            if 'directory' in item:
                recursive = item.get('recursive', False)
                self.update_from_dir(item['directory'], name, action=action, recursive=recursive)
            else:
                self.include_file(item['file'], name, action)

        return self

    def include_file(self, path, name=None, action='update', auto_include=True):
        _dir = self._dir
        log.debug("include_file: path=%s, name=%s, action=%s, _dir=%s", path, name, action, _dir)

        if name is not None:
            if name not in self:
                if action == 'update':
                    self[name] = {}
                elif action == 'append':
                    self[name] = []

            if isinstance(self[name], dict) and not isinstance(self[name], Config):
                self[name] = Config.object(self[name], file=self._file)
            if hasattr(self[name], '_dir'):
                _dir = self[name]._dir
        else:
            assert action != 'append', "append requires a config item name"

        log.debug("include_file: path=%s, name=%s, action=%s, _dir=%s", path, name, action, _dir)
        # if action == 'update':
        #     if name is not None:
        #         _file = self[name].get('_file_')
        #         _dir  = self[name].get('_dir_')

        if not isabs(path):
            path = join(_dir, path)

        log.debug("open %s", path)

        with open(path, 'r') as f:
            for _doc in yaml.load_all(f):
                _doc = Config.object(_doc, file=path)

                if auto_include:
                    _doc.update_from_includes()

                if name is not None:
                    log.debug("%s to self[%s]", action, name)

                    getattr(self[name], action)(_doc)

                    if action == 'update':
                        self[name]._files.add(path)
                        if hasattr(_doc, '_files'):
                            for f in _doc._files:
                                self[name]._files.add(f)

                        # restore _file_ and _dir_
                        # if _file is not None:
                        #     self[name]['_file_'] = _file
                        #     self[name]['_dir_'] = _dir
                        #
                        # if '_files_' not in self[name]:
                        #     self[name]['_files_'] = []
                        #     if _file:
                        #         self[name]['_files_'].append(_file)
                        #
                        # self[name]['_files_'].append(path)
                        #
                else:
                    # _file = self.get('_file_')
                    # _dir  = self.get('_dir_')
                    log.debug("update self")

                    self.update(_doc)

                    self._files.add(path)
                    if hasattr(_doc, '_files'):
                        for f in _doc._files:
                            self._files.add(f)

                    # if _file is not None:
                    #     self['_file_'] = _file
                    #     self['_dir_'] = _dir
                    #
                    # if '_files_' not in self:
                    #     self['_files_'] = []
                    #     if _file:
                    #         self['_files_'].append(_file)
                    #
                    # self['_files_'].append(path)

                    #if '_files_' in _doc:
                    #    self['_files_'] += _doc['_files_']
                    # if '_files_' in self:
                    #     self['_files_'].append(_doc['_file_'])


    def update_from_dir(self, path, name, action='append', recursive=False):
        '''read configuration files and extend config

        Read all yaml files from directory `path` (recursive) and extract all YAML
        documents (also multidocument YAML files) and append to configuration list
        named `name`.
        '''
        _dir = self._dir
        log.debug("update_from_dir:: path=%s, name=%s, action=%s, recursive=%s, _dir=%s", path, name, action, recursive, _dir)

        #import rpdb2 ; rpdb2.start_embedded_debugger('foo')

        if action == 'update':
            if name is not None:
                if not isinstance(self[name], Config):
                    self[name] = Config.object(self[name], file=self._file)

                if self[name]._file == '-':
                    self[name].set_filename(self._file)

                if hasattr(self[name], '_dir'):
                    _dir = self[name]._dir

        if not isabs(path):
            path = join(_dir, path)

        log.debug("  path=%s, exists=%s", path, exists(path))

        if not exists(path): return

        if recursive:
            log.debug("  is_recursive")
            for root, dirs, files in os.walk(path):
                for fn in sorted(files):
                    self.include_file(join(root,fn), name, action)
        else:
            log.debug("  content=%s" % os.listdir(path))
            for fn in sorted(os.listdir(path)):
                _fn = join(path, fn)
                log.debug('  fn=%s, _fn=%s', fn, _fn)
                if isdir(_fn): continue

                log.debug('  is no dir')
                self.include_file(_fn, name, action)

        return self


    def getval(self, name, default=None):
        return self.format(self.get(name, default))


    def format(self, current=Undefined, *data):
        if current is Undefined:
            current = self

        if isinstance(current, string):
            if '{' not in current:
                result = current
            else:
                try:
                    _data = self
                    if data:
                        _data = {}
                        _data.update(self)

                        for d in data:
                            _data.update(d)

                    result = formatter.format(current, **_data)

                    #result = current.format(**_data)
                except KeyError:
                    result = current
                except AttributeError:
                    result = current
            log.debug("format_value result=%s", result)

        elif isinstance(current, (list, tuple)):
            result = [self.format(v) for v in current]

        elif isinstance(current, dict):
            result = {}
            for k,v in current.items():
                result[k] = self.format(v, *(list(data)+[current]))
        else:
            result = current

        return result




"""alerter -- a simple alerter module

"""

from datetime import datetime, timedelta
from dateutil.parser import parse as dt_parse
from itertools import product
from subprocess import Popen, PIPE
from copy import deepcopy
from .notifier import Notifier
from itertools import chain

#from ..config import Config

import logging, sys, json, pyaml, re
log = logging.getLogger('elastico.alerter')

from .util import to_dt, PY3, dt_isoformat, format_value, get_config_value
from .util import stripped, get_alerts

from .config import Config

if PY3:
    unicode = str
    string = str
else:
    string = basestring
    Exception = StandardError

def indent(indent, s):
    if isinstance(indent, int):
        indent = " "*indent
    return "".join([ indent+line for line in s.splitlines(1) ])


class NotificationError(Exception):
    pass

class Alerter:
    '''alerter alerts.

    here more doc.
    '''
    LAST_CHECK = {}
    STATUS     = {}

    @classmethod
    def reset_last_check(cls):
        Alerter.LAST_CHECK = {}

    @classmethod
    def reset_status(cls):
        Alerter.STATUS = {}

    def __init__(self, es_client=None, config={}, config_base="alerter"):
        self.es = es_client
        self.config = config
        self.status_index_dirty = False
        self._refreshed = {}

    def wipe_status_storage(self):
        '''remove all status storages'''
        result = self.es.indices.delete('elastico-alerter-*')
        log.debug("wipe_status_storage: %s", result)
        return result

    def get_status_storage_index(self):
        date = self.now()
        return date.strftime('elastico-alerter-%Y-%m-%d')

    def refresh_status_storage_index(self):
        if self.es:
            try:
                self.es.indices.refresh(self.get_status_storage_index())
            except:
                pass

    def now(self):
        return to_dt(self.config.get('at', datetime.utcnow()))

    def write_status(self, rule, doc_type='elastico_alert_status'):
        storage_type = self.config.get('alerter.status_storage', 'memory')

        #rule['@timestamp'] = to_dt(self.get_rule_value(rule, 'run_at', now))
        rule['@timestamp'] = timestamp = dt_isoformat(self.now())
        if 'at' in rule:
            rule['at'] = dt_isoformat(rule['at'])

        # import socket
        # rule['elastico.hostname'] = socket.gethostname()

        # get data, which is written to any status
        _status = Config(self.config.get('status_data'))
        _status.update(self.config.get('alerter.status_data'))
        rule.update(_status)

        log.debug("rule to write to status: %s", rule)
        log.debug("storage_type=%r", storage_type)

        key  = rule.get('key')
        type = rule.get('type')

        if storage_type == 'elasticsearch':
            index = self.get_status_storage_index()
            #_rule = Config.object(rule).format_value()
            if 'match' in rule and not isinstance(rule['match'], string):
                rule['match'] = json.dumps(rule['match'])
            if 'match_query' in rule and not isinstance(rule['match_query'], string):
                rule['match_query'] = json.dumps(rule['match_query'])

            result = self.es.index(index=index, doc_type=doc_type, body=rule)
            #self.es.indices.refresh(index)
            log.debug("index result: %s", result)
            self.status_index_dirty = True

        elif storage_type == 'filesystem':
            storage_path = self.config.get('alerter.status_storage_path', '')
            assert storage_path, "For status_storage 'filesystem' you must configure 'status_storage_path' "

            path = "{}/{}-{}-latest.yaml".format(storage_path, type, key)
            path = "{}/{}-{}-latest.yaml".format(storage_path, type, key)

            with open(path, 'w') as f:
                json.dump(rule, f)

            # for history
            dt = dt_isoformat(timestamp, '_', 'seconds')
            path = "{}/{}-{}-{}.json".format(storage_path, type, key, dt)
            with open(path, 'w') as f:
                json.dump(rule, f)

        elif storage_type == 'memory':
            if type not in Alerter.STATUS:
                Alerter.STATUS[type] = {}
            Alerter.STATUS[type][key] = rule
            log.debug("set status. type=%r key=%r", type, key)

    def read_status(self, rule=None, key=None, type=None, filter=[]):
        storage_type = self.config.get('alerter.status_storage', 'memory')
        doc_type = 'elastico_alert_status'
        if key is None:
            key  = rule.get('key')
        if type is None:
            type = rule.get('type')

        log.debug("read_status storage_type=%r, key=%r, type=%s", storage_type, key, type)

        if storage_type == 'elasticsearch':
            if self.status_index_dirty:
                self.refresh_status_storage_index()

            results = self.es.search(index="elastico-alerter-*",
                doc_type=doc_type, body={
                'query': {'bool': {'must': [
                    {'term': {'key': key}},
                    {'term': {'type': type}}
                ]+filter}},
                'sort': [{'@timestamp': 'desc'}],
                'size': 1
            })

            if results['hits']['total']:
                result = results['hits']['hits'][0]['_source']
                if 'match' in result:
                    try:
                        result['match'] = json.loads(result['match'])
                    except:
                        pass
                if 'match_query' in result:
                    try:
                        result['match_query'] = json.loads(result['match_query'])
                    except:
                        pass
                return result

            else:
                return None

        elif storage_type == 'filesystem':
            storage_path = self.config.get('alerter.status_storage_path')
            assert storage_path, "For status_storage 'filesystem' you must configure 'status_storage_path' "
            path = "{}/{}-{}-latest.yaml".format(storage_path, type, key)
            with open(path, 'r') as f:
                return json.load(f)

        elif storage_type == 'memory':
            return Alerter.STATUS.get(type, {}).get(key)

    def assert_key(self, data, name=None):
        if name is None:
            name = data.getval('name')
        if 'key' not in data:
            data['key'] = re.sub(r'[^\w]+', '_', name.lower())
        return data.get('key')

    def do_alert(self, alert_data, all_clear=False):
        notifier = Notifier(self.config, alert_data, prefixes=['alerter'])

        # set future status
        # if all_clear:
        #     alert_data['status.current'] = 'ok'
        #     subject = alert_data.getval('subject.ok', '')
        # else:
        #     alert_data['status.current'] = 'alert'
        #     subject = alert_data.getval('subject.alert', '')

        #if isinstance(alert_data.get('message.subject'), string):
        subject = alert_data.getval('message.subject', '')

        if not subject:
            type = alert_data.getval('type')
            name = alert_data.getval('name')
            status  = alert_data['status.current'].upper()
            subject = '[elastico] {} - {} {}'.format(status, type, name)

        # remove_subject = False
        # if 'message.subject' not in alert_data:
        #     remove_subject = True
        #     alert_data['message.subject'] = subject
        #
        log.info("      notification subject=%r", subject)
        notifier.notify(subject=subject)

    def get_query(self, rule, name):
        body = None
        query = rule.getval(name)

        # list of filters
        if isinstance(query, list):
            filters = query

        # lucene query string
        if isinstance(query, string):
            filters = [{'query_string': {'query': query.strip()}}]

        # complete search body (including timerange, if any)
        if isinstance(query, dict):
            return query

        timestamp_field = rule.getval('timestamp_field', '@timestamp')
        timeframe = rule.getval('timeframe', {'minutes': 15})

        if 'endtime' in rule:
            endtime = to_dt(rule.getval('endtime'))
        else:
            endtime = self.now()

        if 'starttime' in rule:
            starttime = to_dt(rule.getval('starttime'))
        else:
            starttime = endtime - timedelta(**timeframe)

        starttime = dt_isoformat(starttime, 'T', 'seconds')#+"Z"
        endtime   = dt_isoformat(endtime, 'T', 'seconds')#+"Z"

        return {
            'query': {'bool': {'must': [
                    {'range': {timestamp_field: {'gte': starttime, 'lte': endtime}}}
                ] + filters
                }},
            'sort': [{timestamp_field: 'desc'}],
            'size': 1
        }


    def _refresh_index(self, index):
        '''make sure index is refreshed at least every 2 minutes'''
        if self.es:
            if index not in self._refreshed:
                self._refreshed[index] = datetime.utcnow() - timedelta(minutes=3)

            if self._refreshed[index] + timedelta(minutes=2) < datetime.utcnow():
                self.es.indices.refresh(index)
                self._refreshed[index] = datetime.utcnow()

                log.info("refreshed index %s", index)


    def do_match(self, rule):

        body = self.get_query(rule, 'match')
        index = rule.get('index')
        body['size'] = 1

        assert index, "index must be present in rule %s" % rule.getval('name')
        rule['match_query'] = body

        key = rule.getval('key')
        type = rule.getval('type')

        if self.es:
            self._refresh_index(index)
            results = self.es.search(index=index, body=body)
            log.debug("results: %s", results)
            rule['match_hits_total'] = results['hits']['total']
            if rule['match_hits_total']:
                rule['match_hit'] = Config.object(results['hits']['hits'][0])
        else:
            rule['match_hits_total'] = 0
            results = {'hits': {'total': 0}}

        # there should be at least min_matches
        min_total = rule.get('matches_min')
        # there should be at most max_matches
        max_total = rule.get('matches_max')
        # ... otherwise we will alert

        # first check if totals are within given bounds
        _result = False
        if min_total is None and max_total is None:
            _result = results['hits']['total'] > 0
        if min_total is not None and max_total is not None:
            _result = results['hits']['total'] >= min_total
            _result = _result and results['hits']['total'] <= max_total

        elif min_total is not None:
            _result = _result or results['hits']['total'] >= min_total

        elif max_total is not None:
            _result = _result or results['hits']['total'] <= max_total

        log.info("match -- key=%r type=%r hits=%r min=%r max=%r trigger=%r "
            "index=%r match_query=%s",
            key, type, results['hits']['total'], min_total, max_total, _result,
            index, json.dumps(rule['match_query']))

        rule['alert_trigger'] = _result
        return _result


    def do_some_command(self, kwargs, rule=None):
        log.debug("do_some_command: kwargs=%s, rule=%s", kwargs, rule)

        if isinstance(kwargs, string):
            kwargs = {'args': kwargs, 'shell': True}
        elif isinstance(kwargs, (list, tuple)):
            kwargs = {'args': kwargs}

        def _get_capture_value(name):
            if name in kwargs:
                return kwargs.pop(name)
            elif rule is not None:
                return rule.get(name)
            else:
                return False
            return

        capture_stdout = _get_capture_value('stdout')
        capture_stderr = _get_capture_value('stderr')

        if 'input' in kwargs:
            input = kwargs.pop('input')
            kwargs['stdin'] = PIPE
        else:
            input = None

        log.debug("run_command: kwargs=%s", kwargs)

        p = Popen(stdout=PIPE, stderr=PIPE, **kwargs)
        (stdout, stderr) = p.communicate(input)
        result = p.wait()

        log.debug("capture_stdout=%s, capture_stderr=%s", capture_stdout, capture_stderr)

        if rule is not None:
            if capture_stdout:
                if stdout.count("\n".encode('utf-8')) == 1:
                    stdout = stdout.strip()
                rule['result.stdout'] = stdout
            if capture_stderr:
                rule['result.stderr'] = stderr
            rule['result.exit_code'] = result

        log.debug("rule: %s", rule)

        return (result, stdout, stderr)

    def do_command_succeeds(self, alert_data):
        cmd = alert_data.getval('command_succeeds')
        (result, stdout, stderr) = self.do_some_command(cmd, alert_data)

        expect_code = alert_data.get('expect.code', 0)

        log.info("msg='command_succeeds' "
            "cmd=%r result=%r expect=%r stdout=%r stderr=%r",
            cmd, result, expect_code, stripped(stdout), stripped(stderr))

        _result = not (result == expect_code)
        alert_data['alert_trigger'] = _result
        return _result

    def do_command_fails(self, alert_data):
        cmd = alert_data.get('command_fails')
        (result, stdout, stderr) = self.do_some_command(cmd, alert_data)

        expect_code = alert_data.get('expect.code', 0)

        log.info("msg='command_fails' "
            "cmd=%r result=%r expect=%r stdout=%r stderr=%r",
            cmd, result, expect_code, stripped(stdout), stripped(stderr))

        _result = not (result != expect_code)
        alert_data['alert_trigger'] = _result
        return _result

    def check_alert(self, alert_data, status=None):
        if not isinstance(alert_data, Config):
            alert_data = Config(alert_data)

        _key = alert_data.get('key')

        logger_name = alert_data.getval('logger', 'elastico.alerter.%s' % _key)
        log = logging.getLogger(logger_name)

        if status is None:
            # get last status of this alert_data
            last_rule = self.read_status(alert_data)

            if last_rule is not None:
                if isinstance(last_rule['status'], string):
                    status = last_rule['status']
                else:
                    status = last_rule['status']['current']
            log.debug("current_status=%r", last_rule)

        if status is None:
            alert_data['status.previous'] = 'ok'
            alert_data['status.current']  = 'ok'
        else:
            alert_data['status.previous'] = status
            alert_data['status.current'] = status

        need_alert = False
        if 'command_fails' in alert_data:
            need_alert = need_alert or self.do_command_fails(alert_data)

        if 'command_succeeds' in alert_data:
            need_alert = need_alert or self.do_command_succeeds(alert_data)

        if 'match' in alert_data:
            need_alert = need_alert or self.do_match(alert_data)

        if not need_alert:
            alert_data['status.current'] = 'ok'
        else:
            alert_data['status.current'] = 'alert'

            log.warning("need alert -- name=%r, status=%r", alert_data.getval('name'), status)
            # new status = alert
            if status != 'ok' and last_rule:
                realert = alert_data.get('realert', {'minutes': 60})

                if realert == 'never':
                    alert_data['status.realert'] = 'never'
                    return alert_data
                # else it is assumed, that realert is a dictionary, which can
                # be passed to timedelta

                # calculate wait time till next re-alert
                now = self.now()
                try:
                    delta = timedelta(seconds=int(last_rule['status']['realert']))
                except Exception as e:
                    log.error("error", exc_info=1)
                    delta = timedelta(**realert)

                time_passed = now - to_dt(last_rule['@timestamp'])

                log.info("delta=%r time_passed=%r", delta, time_passed)

                # if there is still wait time left,
                if delta > time_passed:
                    wait_time = delta - time_passed
                    log.info("wait_time=%r", wait_time)

                    #alert_data['status.realert'] = 'wait'
                    alert_data['status.realert'] = wait_time.total_seconds()
                    log.warning("      trigger alert -> wait for realert (%s)", wait_time)
                    return alert_data

                else:
                    if alert_data['status.previous'] != 'ok':
                        alert_data['status.current'] = 'realert'
                    log.warning("      trigger alert -> %s", alert_data['status.current'])

###>
#            log.info("      trigger alert")
#
#            self.do_alert(alert_data)
#
#        else:
#            if status == 'alert':
#                log.info("      trigger ok")
#                self.do_alert(alert_data, all_clear=last_rule)
#            else:
#                log.info("      trigger nothing")
#
#        if not alert_data.get('dry_run'):
#            self.write_status(alert_data)
#
#        # here we can expand everything
#        # check result and log
###<
        return alert_data

    def process_rules(self, action=None, **arguments):
        self._refreshed = {}
        if 'arguments' not in self.config:
            self.config['arguments'] = {}

        self.config['arguments'].update(arguments)
        rules = self.config.get('alerter.rules', [])
        log.debug("rules: %s", rules)

        # # sort rules first
        # ordered_rules = []
        # for rule in rules:
        #      rule.get('depends')

        for rule in rules:
            if not rule: continue

            log.debug("rule: %s", rule)

            if isinstance(rule, string):
                rule = rules[rule]

            rule = self.config.assimilate(rule)
            # TODO: why is assimilate needed here?  should already be done

            _class_name = rule.getval('class')
            if _class_name is None:
                _class_name = rule.getval('name')

            log.debug("rule: %s", rule)
            log.info("=== RULE <%s> =========================", _class_name)

            self.process(rule, action=action)

        if self.es:
            self.refresh_status_storage_index()

    ALIAS = re.compile(r"^\s*\*(\w+)(\.\w+)*(\s+\*(\w+)(\.\w+)*)*\s*$")
    def process(self, rule, action=None):
        log = globals()['log']

        has_foreach = False

        now = self.now()

        # create a product of all items in 'each' to multiply the rule
        if 'foreach' in rule:
            data_list = []

            for key,val in rule.get('foreach', {}).items():
                log.debug("key: %s, val: %s", key, val)

                # maybe format the values
                key = rule.format(key)

                # expand *foo.bar values.
                if isinstance(val, string):
                    val = rule.format(val)

                    log.debug("val is aliases candidate")
                    _value = []
                    if self.ALIAS.match(val):
                        log.debug("there are aliases: %s", val)

                        _refs = val.strip().split()
                        for _ref in _refs:
                            _val = rule.get(_ref[1:])

                            if _val is None:
                                _val = self.config.get(_ref[1:])

                            assert _val is not None, "could not resolve reference %s mentioned in rule %s" % (_ref, rule['name'])
                            _value += _val

                        val = _value
                    else:
                        log.debug("no aliases: %s", val)

                #if val == '@'
                data_list.append([{key: rule.format(v)} for v in val])

            data_sets = product(*data_list)

            has_foreach = True

        else:
            data_sets = [({},)]

        visited_keys = []


        for data_set in data_sets:
            log.debug("data_set: %s", data_set)

            r = Config.object()

            # # copy config's data section
            # r.update(deepcopy(self.config.get('data', {})))
            #
            # # copy alerter's data section
            # r.update(deepcopy(self.config.get('alerter.data', {})))

            # copy arguments
            r.update(deepcopy(self.config.get('arguments', {})))

            # for k,v in self.config.items():
            #     if k.endswith('_defaults'): continue
            #     if k in ('arguments', 'alerter', 'digester'): continue
            #     r[k] = deepcopy(v)

            # get defaults
            defaults = self.config.get('alerter.rule_defaults', {})
            _class = rule.get('class', 'default')

            log.debug("rule class: %s", _class)
            _defaults = defaults.get(_class, {})

            log.debug("rule defaults: %s", _defaults)
            r.update(deepcopy(_defaults))

            # update data from rule
            r.update(deepcopy(rule))

            for data in data_set:
                r.update(deepcopy(data))

            _name = r.getval('name')
            _key = self.assert_key(r, _name)

            # overrides from arguments
            #r.update(self.config.get('alerter.rule.%s', {}))

            logger_name = r.getval('logger', 'elastico.alerter.%s' % _key)
            log = logging.getLogger(logger_name)

            log.info("rule=%r logger=%r", _name, logger_name)

            _alerts = get_alerts(r.get('alerts', []), context=r)

            collected_alerts = []

            for alert in _alerts:
                log.debug("process alert %s", alert)
                alert_data = Config.object()

                assert 'type' in alert

                _type = alert_data.format(alert['type'])

                defaults = self.config.get('alerter.alert_defaults', {})
                alert_data.update(deepcopy(defaults.get(_type,{})))

                defaults = r.get('alert_defaults', {})
                alert_data.update(deepcopy(defaults.get(_type,{})))

                alert_data.update(r)

                alert_data.update(alert)
                alert_data = alert_data.format()

                if 'alerts' in alert_data:
                    del alert_data['alerts']

                log.debug("alert_data (alert): %s", alert_data)

                _r_name = r.getval('name')
                _key = self.assert_key(alert_data, _r_name)

                log.info("alert? -- key=%r, type=%r", _key, _type)

                visit_key = (_type, _key)
                assert visit_key not in visited_keys, \
                    "key %s already used in rule %s" % (_key, _r_name)

                assert 'match' in alert_data or 'no_match' in alert_data \
                    or 'command_succeeds' in alert_data \
                    or 'command_fails' in alert_data, \
                    "rule %s does not have a check defined" % _r_name

                log.debug("alert_data: %s", alert_data)

                # check only every now and then. default 2min
                every = timedelta(**alert_data.get('every', {'minutes': 1}))
                #log.debug("LAST_CHECK=%r", Alerter.LAST_CHECK)
                last_check = Alerter.LAST_CHECK.get(visit_key, now - every - timedelta(seconds=1))
                log.debug("last_check=%s, now=%s, every=%s", last_check, now, every)

                if (now - every) >= last_check:
                    # if action:
                    #     action(Config(alert_data))
                    # else:
                    collected_alerts.append(self.check_alert(alert_data))
                    Alerter.LAST_CHECK[visit_key] = now
                else:
                    log.info("      next check in %s", every - (now-last_check))

            if 1:
                # helper function to always return a list
                def _get_list(D,n):

                    l = D.get(n, [])
                    if not isinstance(l, list):
                        l = [l]
                    return l

                log.debug("rule: %r", r)

                self.assert_key(r)

                # get rule status
                rule_status = self.read_status(type='rule', key=r.getval('key'))

                if rule_status is None:
                    rule_status = {}

                rule_status['key'] = r.getval('key')
                rule_status['type'] = 'rule'
                rule_status['name'] = r.getval('name')

                alerts = collected_alerts
                log.debug("alert count=%s", len(alerts))

                #
                was_ok = all([ a['status.previous'] == 'ok' for a in alerts ])
                now_ok = all([ a['status.current'] == 'ok' for a in alerts ])
                was_alert = lambda a: a['status.previous'] != 'ok'
                was_alert = [ x for x in filter(was_alert, alerts)]
                log.debug("was_ok=%r now_ok=%r was_alert=%r", was_ok, now_ok, was_alert)

                # update rule status
                if was_ok and not now_ok or 'status.start' not in rule_status:
                    rule_status['status.start'] = dt_isoformat(now)


                # all related alerts and records will get this id
                rule_status['status.id'] = '{}_{}'.format(
                    rule_status['key'], rule_status['status.start'])

                # notify
                if was_ok:
                    notify = []
                else:
                    notify = _get_list(rule_status, 'notify')

                notify_lists = [_get_list(a, 'notify') for a in was_alert]
                notify = list(set(chain(notify, *notify_lists)))
                rule_status['notify'] = notify

                # alerts
                if was_ok:
                    al_list = []
                else:
                    al_list = _get_list(rule_status, 'alerts')
                rule_status['alerts'] = list(set(chain( al_list,
                    [ a['type'] for a in was_alert ]
                )))

                if was_alert and now_ok:
                    rule_status['status.end'] = dt_isoformat(now)

                    all_clear = Config.object()
                    all_clear.update(r)
                    all_clear['type'] = 'all-clear'
                    all_clear['status.current'] = 'ok'
                    all_clear['notify'] = rule_status['notify']
                    self.assert_key(all_clear)
                    all_clear.update(r.get('all_clear', {}))
                    log.info("all_clear=%r", all_clear)
                    #log.info("all_clear=%r", all_clear)
                    self.do_alert(all_clear)

                    for alert in alerts:
                        alert['status.start'] = rule_status['status.start']
                        alert['status.id']    = rule_status['status.id']
                        alert['status.end']   = rule_status['status.end']

                        if not rule.get('dry_run'):
                            self.write_status(alert)
                else:
                    log.debug("iter alerts")
                    for alert in alerts:
                        alert['status.start'] = rule_status['status.start']
                        alert['status.id']    = rule_status['status.id']

                        log.debug("alert=%r", alert)
                        if action:
                            action(alert)
                        else:
                            status_current = alert['status.current']
                            status_realert = alert.get('status.realert')

                            if status_current != 'ok' and not status_realert:
                                self.do_alert(alert)

                            if not rule.get('dry_run'):
                                self.write_status(alert)

                    # send all clear to all notifications in rule_status


                if not r.get('dry_run'):
                    self.write_status(rule_status)
                        #doc_type='elastico_rule_status')

                # finally write status


#
#            all_clear = True
#            for alert in collected_alerts:
#               if alert['status.current'] == 'ok':
#               elif 'realert' not in alert['status']:
#                  self.do_alert(alert)
#               all_clear = all_clear and alert['status.current'] == 'ok'
#
#            raise all clear only if all alerts are gone
#

        # if not alert_data.get('dry_run'):
        #     self.write_status(alert_data)

            if not _alerts and action:
                action(rule)

    @classmethod
    def run(cls, config):
        '''run alerter
        '''

        from .connection import elasticsearch
        es = elasticsearch(config)

        sleep_seconds = config.get('sleep_seconds')
        alerter = Alerter(es, config)
        if sleep_seconds:
            while True:
                try:
                    alerter.process_rules()
                    time.sleep(sleep_seconds)
                except Exception as e:
                    log.error("exception occured while processing rules: %s", e)
        else:
            alerter.process_rules()

    @classmethod
    def expand_rules(cls, config):
        '''expand alert rules
        '''
        RULES = []
        def collect_rules(rule):
            RULES.append(rule.format())
            return rule

        Alerter(None, config).process_rules(action=collect_rules)
        return RULES




"""cli.alerter -- control alerter

With ``alerter`` command you can control the :py:mod:`~elastico.alerter`
module.

For more help on a command, run::

   elastico alerter <command> -h

"""
from .cli import command, opt, arg
from ..alerter import Alerter
from ..connection import elasticsearch
from ..util import write_output
from ..server import Server

import pyaml, logging, time
logger = logging.getLogger('elastico.cli.alerter')

alerter_command = command.add_subcommands('alerter', description=__doc__)

@alerter_command("expand-rules",
    arg("--list", '-l', choices=['names', 'keys', 'types', 'alerts'], default=None),
    arg("--format", '-f', default=None),
    )
def alerter_expand_rules(config):
    """Expand rules, that you can check, if they are correct

    This command expands the rules like in a regular alerter run and prints
    them to stdout in YAML format.  This way you can check, if all variables
    and defaults are expanded as expected.
    """
    expanded_rules = Alerter.expand_rules(config)
    if config['alerter.expand-rules.list']:
        expand = config['alerter.expand-rules.list']

        if expand in ('names', 'keys', 'types'):
            for name in set([ rule[expand[:-1]] for rule in expanded_rules ]):
                print(name)

        if expand == 'alerts':
            for name in set([ "%s-%s" % (rule['type'], rule['key']) for rule in expanded_rules ]):
                print(name)

    elif config['alerter.expand-rules.format']:
        for rule in expanded_rules:
            print(config['alerter.expand-rules.format'].format(**rule))
    else:
        pyaml.p(expanded_rules)

@alerter_command('check',
    arg('--status', "-s", choices=['ok', 'alert', 'error'], default='ok'),
    arg('alert', nargs="*", default=[]),
    )
def alerter_check(config):

    config['arguments.dry_run'] = True

    result = []
    alerter = Alerter(elasticsearch(config), config)
    check_alerts = config.get('alerter.check.alert')
    status = config['alerter.check.status']

    def check(alert):
        logger.debug("alert: %s", alert)

        alert_id = "%s-%s" % (alert['type'], alert['key'])
        if (check_alerts
            and alert_id not in check_alerts
            and alert['key'] not in check_alerts): return

        result.append(alerter.check_alert(alert, status=status))

    alerter.process_rules(action=check)

    write_output(config, result)

@alerter_command("run")
def alerter_run(config):
    """run alerter"""
    alerter = Alerter(elasticsearch(config), config)
    alerter.process_rules()

@alerter_command("serve",
    arg('--sleep-seconds', '-s', type=float, default=60, config="serve.sleep_seconds"),
    arg('--count', '-c', type=int, default=0, config="serve.count"),
    )
def alerter_serve(config):
    """run alerter"""

    def _run():
        alerter = Alerter(elasticsearch(config), config)
        alerter.process_rules()

    server = Server(config, run=_run)
    server.run()

@alerter_command("query")
def alerter_run(config):
    """run alerter"""
    pass



'''idea of digest commands is to analyse and do digestions.

Here an example for aggregating filesystem metricset of metric beat:

Configuration:
```yaml
# for each index matching index pattern
retentions:
- index-pattern: metricbeat-*

  target: history-metricbeat-%Y-%m

  digest:
  - name: metricset-filesystem

    fields:
        metricset.module: "system"
        metricset.name: "filesystem"

    buckets:
        terms:
            tags:        tags
            host_name:   host.name
            mount_point: system.filesystem.mount_point

    aggregate:
        terms:
            metricset.name:   metricset.name
            metricset.module: metricset.module
        stats:
            available:  system.filesystem.available
            free:       system.filesystem.free
            total:      system.filesystem.total
            available:  system.filesystem.available
            used_pct:   system.filesystem.used.pct
            used_bytes: system.filesystem.used.bytes

  - name: metricset-network

      fields:
        metricset.module: "system"
        metricset.name: "filesystem"

      buckets:
        terms:
          - host.name
          - system.network.name

      aggregate:
        terms:
            - tags
        stats:
            - system.network.in.bytes
            - system.network.in.dropped
            - system.network.in.errors
            - system.network.in.packats

            - system.network.out.bytes
            - system.network.out.dropped
            - system.network.out.errors
            - system.network.out.packats
```

```json
```


'''
from .cli import command

digester_command = command.add_subcommands('digester',)

digester_command('query',
)



from copy import deepcopy
from textwrap import indent
from .config import Config
from .util import run_command, string, PY3
from pprint import pformat

if PY3:
    unicode = str

import logging, pyaml
log = logging.getLogger('elastico.notifier')

class BaseNotifier:
    def __init__(self, config, data=None, prefixes=[]):
        assert isinstance(config, Config), "config must be Config instance"
        if data is None:
            data = Config.object()

        assert isinstance(data, Config), "data must be Config instance"

        self.config = config
        self.data = data
        self.prefixes = prefixes

    def update_data(self, data, value=None):
        if value is None:
            self.data.update(data)
        else:
            self.data[data] = value

    def get_value(self, name, default=None, configs=None):
        for cfg in configs:
            try:
                return cfg[name]
            except KeyError:
                pass
        try:
            return self.data[name]
        except KeyError:
            pass

        try:
            return self.config[name]
        except KeyError:
            return default






class EmailNotifier(BaseNotifier):

    def notify(self, message, *configs):
        _get = lambda n,d=None: self.get_value(n,d,configs=configs)

        smtp_host    = _get('smtp.host', 'localhost')
        smtp_ssl     = _get('smtp.ssl', False)
        smtp_port    = _get('smtp.port', 0)

        email_from   = _get('email.from', 'noreply')
        email_cc     = _get('email.cc', [])
        email_to     = _get('email.to', [])
        email_bcc    = _get('email.bcc', [])

        if not isinstance(email_cc, list) : email_cc  = [email_cc]
        if not isinstance(email_to, list) : email_to  = [email_to]
        if not isinstance(email_bcc, list): email_bcc = [email_bcc]

        recipients = email_to + email_cc + email_bcc

        assert recipients, "you must specify email recipient"

        from email.mime.multipart import MIMEMultipart
        from email.mime.text import MIMEText

        (plain, html) = (None, None)

        if message.get('plain'):
            plain = MIMEText(message['plain'], 'plain')
        if message.get('html'):
            html = MIMEText(message['html'], 'html')

        if plain and html:
            msg = MIMEMultipart('alternative')
            msg.attach(plain)
            msg.attach(html)
        elif plain:
            msg = plain
        elif html:
            msg = html

        def _set_email_header(key, value):
            log.info("alert_email: %s: %s", key, value)
            if isinstance(value, list):
                msg[key] = ", ".join(value)
            else:
                msg[key] = value

            self.update_data({'email.%s' % key.lower(): msg[key]})

        _set_email_header('From', email_from)
        _set_email_header('Subject', message['subject'])
        _set_email_header('To', email_to)

        if email_cc:
            _set_email_header('Cc', email_cc)

        log.info("alert_email: Bcc: %s", email_bcc)

        email_message = msg.as_string()

        from .util import sendmail

        sendmail_result = sendmail(
                host=smtp_host,
                port=smtp_port,
                use_ssl=smtp_ssl,
                username=None,
                password=None,
                sender=email_from,
                recipients=recipients,
                message=email_message
                )

        result = {}
        if sendmail_result:
            for recipient in recipients:
                if recipient not in sendmail_result:
                    result[recipient] = {'status': 200, 'message': 'ok'}
                else:
                    status, msg = result[recipient]
                    result[recipient] = {'status': status, 'message': msg}

            raise NotificationError("Some recipients had errors", result)

        return sendmail_result


class CommandNotifier(BaseNotifier):
    def notify(self, message, *configs):
        _get = lambda n,d=None: self.get_value(n,d,configs=configs)

        cmd = _get('command')
        cmd = self.data.format(cmd, message)

        if not self.config.get('dry_run'):
            (result, stdout, stderr) = run_command(cmd, self.data)


class Notifier(BaseNotifier):
    transports = {
        'email': EmailNotifier,
        'command': CommandNotifier,
    }

    def compose_message_text(self, message, rule, **kwargs):
        '''compose message text from text with data from alert and rule
        '''

        log.debug("compose_message_text message=%r, rule=%r, kwargs=%r",
            message, rule, kwargs)

        if isinstance(message, string):
            message = {'text': message}

        import markdown

        try:
            data  = indent(pyaml.dump(rule, dst=unicode), " "*4)+"\n"
        except Exception as e:
            log.error("could not convert data to YAML -- message='%s', rule=%r", e, rule)

            data = "Could not convert data to YAML.\n\n"
            data += indent(pformat(rule)," "*4)

        try:
            #if message
            plain = message.get('plain', '{message.text}\n---------\n\n{message.data}')
            text  = message.get('text', '')
            text  = rule.format(text, Config(kwargs))
            plain = rule.format(plain, Config(kwargs), Config({'message': {'data': data, 'text': text}}))
            html  = markdown.markdown(plain)
        except Exception as e:
            import traceback
            log.error("could not compose message -- message=%r, kwargs=%r", e, kwargs)
            text = "Could not compose text or plain message:\n\n"
            text += traceback.format_exc()
            text += "\n\n"+data

            plain = text
            html = None

        return (text, data, plain, html)

    def transport_notification(self, message, notify_spec, data):
        notify_class = Notifier.transports[notify_spec['transport']]
        notifier = notify_class(self.config, notify_spec)
        notifier.notify(message, notify_spec, data)

    def notify(self, notify=None, data=None, subject=None, text=None):
        if data is None:
            data = self.data
        if notify is None:
            notify = data.get('notify', [])

        # compose predefined notifications
        notify_specs = self.config.get('notifications', {})
        for prefix in self.prefixes:
            notify_specs.update(self.config.get('%s.notifications' % prefix, {}))

        # normalize notify data
        if isinstance(notify, dict):
            _tmp = []
            for k,v in notify.items():
                _notification = deepcopy(v)
                _notification['notification'] = k
                _tmp.append(_notification)
            notify = _tmp

        notifications = {}
        for notify_name in notify:
            try:
                notify_spec = Config.object()

                if isinstance(notify_name, string):
                    notify_spec.update(deepcopy(notify_specs[notify_name]))
                    notify_spec['notification'] = notify_name

                else:
                    notify_spec.update(deepcopy(notify_name))
                    notify_name = notify_spec['notification']

                underscore = Config(data.get('match_hit._source', {}))

                text, data_s, plain, html = self.compose_message_text(
                    data.get('message', {}),
                    data,
                    _ = underscore,
                    )

                log.debug("data: %r", data)

                if subject is None:
                    subject = data.getval('message.subject',
                        '[elastico] notification')

                message = {
                    'text': text,
                    'data': data_s,
                    'plain': plain,
                    'html': html,
                    'subject': subject,
                }

                notify_spec['message.subject'] = subject
                notify_spec['message.text'] = text

                self.transport_notification(message, notify_spec, data.format())

                if self.config.get('dry_run'):
                    notify_spec['status'] = 'dry_run'
                else:
                    notify_spec['status'] = 'ok'

                notifications[notify_name] = data.format(notify_spec)

            except Exception as e:
                # log.error('Error while processing notification %s: %s', notify_name, e)

                notify_spec['status'] = 'error'

                args = e.args[1:]
                if len(args) > 1:
                    details = dict( (str(i), a) for a in enumerate(args, 1)  )
                elif len(args) == 1:
                    details = args[0]
                if len(args) == 0:
                    details = None

                if hasattr(e, 'message'):
                    message = e.message
                else:
                    message = e.__class__.__name__+"("+str(e)+")"

                log.error("      notification error %s", message, exc_info=1)
                notify_spec['error'] = {
                    'message': message,
                    'details': details,
                }

                log.debug('nspec[error]: %s', notify_spec['error'])

            log.info("      notification %s -> %s", notify_name, notify_spec['status'])

            self.update_data('notifications', notifications)




import logging, requests

from argdeco import command, main, arg, opt, config_factory
from os.path import exists, join, expanduser
from zipfile import ZipFile, ZipInfo

from ..util import to_dt, dt_isoformat, write_output
from ..config import Config

import os
import logging
from datetime import datetime

# initialize logger
logger = logging.getLogger('elastico.cli')
import os

# find config_file default
# ------------------------
if 'USERPROFILE' in os.environ and 'HOME' not in os.environ:
    os.environ['HOME'] = os.environ['USERPROFILE']

config_file_user   = expanduser("~/.config/elastico/elastico.yml")
config_file_system = "/etc/elastico/elastico.yml"

if exists(config_file_user):
    config_file_default = config_file_user
elif exists(config_file_system):
    config_file_default = config_file_system
else:
    config_file_default = None

# configure main function
# -----------------------
main.configure(
    compiler_factory=config_factory(Config,
        prefix = 'arguments',
        config_file=arg( '--config-file', '-C',
            help="File to read configuration from",
            default = config_file_default,
            ),
    ),
    debug=True,
    verbosity=True,
    prog="elastico",
    )

@arg('--at',
    help="simulate running this program at given time",
    config="at",
    default=dt_isoformat(datetime.utcnow(), 'T', 'seconds')
    )
def arg_at(value):
    return dt_isoformat(to_dt(value))

# add global arguments
# --------------------
main.add_arguments(
    arg('--host', '-H', help="Elasticsearch host. (CFG: elasticsearch.hosts)", config="elasticsearch.hosts"),
    arg('--netrc', help="get netrc entry <machine>. (CFG: netrc.machine)", config="netrc.machine"),
    arg('--netrc-file', help="set netrc file to read. (CFG: netrc.file)", config="netrc.file"),
    arg('--output-format', '-F', help="set output format", choices=['json', 'yaml'], default='yaml'),
    arg_at,
)

class MyZipFile(ZipFile):
    '''This class overrides :py:meth:~zipfile.ZipFile:'s ``_extract_member``
    method to set executable flags, if set in ZIP file'''

    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, ZipInfo):
            member = self.getinfo(member)

        if targetpath is None:
            targetpath = os.getcwd()

        ret_val = ZipFile._extract_member(self, member, targetpath, pwd)
        attr = member.external_attr >> 16
        os.chmod(ret_val, attr)
        return ret_val


@command('install')
def install(config):
    '''install elasticsearch from zip right here, mostly used for testing'''

    package = "elasticsearch"
    version = "6.3.2"
    filename = "{}-{}.zip".format(package, version)

    if not exists(filename):
        url = "https://artifacts.elastic.co/downloads/elasticsearch/"+filename

        logger.info("Downloading %s", url)

        r = requests.get(url)
        with open(filename, 'wb') as output:
            for chunk in r.iter_content(chunk_size=512*1024):
                output.write(chunk)

    logger.info("Extracting %s", filename)
    import zipfile
    zip_ref = MyZipFile(filename, 'r')
    zip_ref.extractall(".")
    zip_ref.close()

@command('run')
def install(config):
    """run elastic search in current directory"""

    package = "elasticsearch"
    version = "6.3.2"

    executable = "{}-{}/bin/elasticsearch".format(package, version)

    #if not exists(executable)

    os.execl(executable, executable)

@command('check-args')
def check_args(config):
    """check config"""
    import pyaml
    pyaml.p(config)




from .util import to_dt, dt_isoformat
from .util import string
from datetime import datetime, timedelta

def search(es, query=None, index=None):
    return es.search(index=index or '*', body=build_query_body(query))

def build_query_body(query):
    return {'query': {'query_string': {'query': query}}}

def build_search_body(config, name):
    '''build a search body from given key `name` in config.

    '''
    body = None
    # list of filters
    if isinstance(config[name], list):
        filters = config[name]

    # lucene query string
    if isinstance(config[name], string):
        filters = [{'query': {'query_string': {'query': config[name]}}}]

    # complete search body (including timerange, if any)
    if isinstance(config[name], dict):
        return config[name]

    timestamp_field = config.get('timestamp_field', '@timestamp')
    timeframe = config.get('timeframe_minutes', 60)

    if 'endtime' in config:
        endtime = to_dt(config['endtime'])
    else:
        endtime = datetime.utcnow() #.isoformat('T', 'seconds')+"Z"

    if 'starttime' in config:
        starttime = to_dt(config['starttime'])
    else:
        starttime = endtime - timedelta(minutes=timeframe)

    starttime = dt_isoformat(starttime)
    endtime   = dt_isoformat(endtime)

    return {
        'query': {'bool': {'must': [
                {'range': {timestamp_field: {'gte': starttime, 'lte': endtime}}}
            ] + filters,
        'sort': {timestamp_field: 'desc'}
    }}}



class DataProcessor(object):
    expand_config_items = []
    config_item = None

    def __init__(self, config, es_client=None):
        self.config    = config
        self.es = es_client
        self.status = {}

    def __getattr__(self, name):
        if name == 'bulk_docs':
            self.bulk_docs = []
            return self.bulk_docs
        return DataProcessor.__getattr__(self, name)

    def bulk_append(self, doc, docs=None):
        '''append a document to bulk buffer

        Flushing happens automatically, when there are `bulk.size` elements in
        bulk buffer.
        '''
        if docs is None:
            docs = self.bulk_docs

        assert '_id' in doc, "you have to set _id field"
        assert '_index' in doc, "you have to set _index field"

        docs.append(doc)
        size = self.config.get("bulk.size", 1000)
        if len(docs) > size:
            return self.bulk_flush(docs)

        return (0, [])

    def bulk_flush(self, docs=None):
        '''flush bulk buffer to elasticsearch
        '''
        if docs is None:
            docs = self.bulk_docs

        from elasticsearch.helpers import bulk
        result = bulk(self.es, docs)
        self.es.indices.refresh('_all')
        docs[:] = []
        return result

    def process(self, rule, action=None):
        pass

    def get_required(self, item, name, message=None):
        value = item.get(name)
        if message is None:
            message = "{name} required in {item}"
        assert value is not None, message.format(item=item['name'], name=name)
        return value

    def process_rules(self, action=None):
        rules = self.config.get("%s.rules" % self.config_item, [])
        for i,rule in enumerate(rules, start=1):
            rule.update_from_includes()
            rule_name = rule.get('name')
            assert rule_name, "you must provide a name for rule #%s from file %s" % (i, self.config.get_filename(rule))
            self.process(rule, action=action)




from .cli import command, arg
from os.path import exists
from ..search import build_search_body
from ..connection import elasticsearch

import logging
log = logging.getLogger('elastico.cli.search')

@command('search',
    arg('--format', '-F',
        help="format string, which is applied to each match",
        default=None),
    arg('query', help="may be a query, a filename or '-' to read from stdin"),
)
def search(config): #host, format, query):
    format = config.get('search.format')
    query  = config.get('search.query')

    log.debug("format=%r query=%r", format, query)
    content = None
    if query == '-':
        content = sys.stdin.read()
    elif exists(query):
        with open(query, 'r') as f:
            content = f.read()
    else:
        config = {}
        config['query'] = query
        config['format'] = format

    if content is not None:
        if content.startswith("---"):
            import frontmatter
            config, content = frontmatter.loads(content)
            config['search.template'] = content
        else:
            config = yaml.loads(io.StringIO(content))

    body = build_search_body(config, 'query')
    log.error("body=%r", body)
    es = elasticsearch(config)
    results = es.search(index=config.get('index', '*'), body=body)

    if config['search.format']:
        for hit in results['hits']['hits']:
            print(config['search.format'].format(hit))

    elif config['search.template']:
        import chevron
        args = {}
        for k in ('template', 'partials_path', 'partials_ext',
           'partials_dict', 'padding', 'def_ldel', 'def_rdel'):
            if k in config['search']:
                args[k] = config['search'][k]

        data = config.get('data', {}).copy()
        data.update(results)
        args['data'] = data
        print(chevron.render(**args))

    else:
        import pyaml
        pyaml.p(results)




from .cli import command, arg

index_command = command.add_subcommands('index',help="work with indices")

@index_command('ls', arg('index', help="index name(s)", nargs="*"))
def cmd_indices(config):
    '''list indices'''
    from ..connection import elasticsearch
    es = elasticsearch(config)
    indexes = config.get('index.ls.index')
    if not indexes:
        indexes = ['_all']
    for indexpattern in indexes:
        for idx in es.indices.get(indexpattern).keys():
            print(idx)

@index_command('rm', arg('index', help="index name(s)", nargs="+"))
def cmd_indices(config):
    '''list indices'''
    from ..connection import elasticsearch
    es = elasticsearch(config)
    for idx in config.get('index.rm.index'):
        es.indices.delete(idx)

@index_command('refresh', arg('index', help="index name(s)", nargs="+"))
def cmd_indices(config):
    '''list indices'''
    from ..connection import elasticsearch
    es = elasticsearch(config)
    for idx in config.get('index.refresh.index'):
        print(es.indices.refresh(idx))

    # for idx in es.indices.get('_all').keys():
    # for idx in es.indices.get('_all').keys():
    #     print(idx)



"""digest data for having a history

Having e.g. metricbeat data from every 15 seconds or every minute results in
a lot of detailed data.  Looking back in history it is not so important for
having all of the data, but rather having averages, minima, maxima, etc.

This module shall provide tools for easy aggregating data (maybe even
automatic) and put them into buckets and write them to some history index.

"""

from .util import string, to_dt, start_of_day, end_of_day, first_value

import logging
log = logging.getLogger('elastico.digest')

from .data_processor import DataProcessor
from datetime import datetime, timedelta

class Digester(DataProcessor):
    config_item = 'digest'


    def get_aggregate_items(self, agg_type, field_specs):
        aggs = []

        # type 1: dictionary
        if isinstance(field_specs, dict):

            # item type 1b: implicit name -> field mapping
            for name, field in field_specs.items():
                aggs.append({name: {agg_type: {"field": field}}})
        else:
            # type 2: list
            for field_spec in field_specs:
                # item type 1: dict
                if isinstance(field_spec, dict):
                    # item type 1a: explicit definition
                    if 'field' in field_spec:
                        name  = field_spec['name']
                        field = field_spec['field']
                        _field_spec = {"field": field}
                        for k,v in field_spec:
                            if k in ('name', 'field'): continue
                            _field_spec[k] = v
                        aggs.append((name, {agg_type: _field_spec}))
                    else:
                        # item type 1b: implicit name -> field mapping
                        for k,v in field_spec.items():
                            aggs.append((k, {agg_type: {"field": v}}))
                # item type 2: string:   string -> string mapping
                else:
                    aggs.append({field_spec: {agg_type: {"field": field_spec}}})
        return aggs


    def make_range_query(self, min_date=None, max_date=None):
        range = None
        if min_date or max_date:
            range = {"range": {timestamp_field: {}}}

        if min_date:
            range['range'][timestamp_field]['gte'] = dt_isoformat(min_date)

        if max_date:
            range['range'][timestamp_filed]['lte'] = dt_isoformat(max_date)

        return range


    def make_query(self, rule, digest, min_date=None, max_date=None):
        import pprint
        # dictionary of final aggregations
        aggs = {}
        for agg_type, field_specs in digest.get('aggregates', {}).items():
            agg_items = self.get_aggregate_items(agg_type, field_specs)
            for agg_item in agg_items:
                aggs.update(agg_item)

        log.debug("aggs: %s", aggs)

        # setup the buckets
        for bucket_type, field_specs in digest.get('buckets', {}).items():
            agg_items = self.get_aggregate_items(bucket_type, field_specs)
            log.debug("bucket - agg_items: %s", agg_items)
            for spec in agg_items:
                value = first_value(spec)
                value['aggs'] = aggs
                aggs = spec

        log.debug("aggs: %s", aggs)

        # setup query
        query = digest.get('query')

        # setup condition
        if isinstance(query, list):
            query = {'bool': {'must': query}}
        elif isinstance(query, string): # string
            query = {'bool': {'must': [{'query_string': {'query': query}}]}}

        range = self.make_range_query(min_date, max_date)
        if range:
            query['bool']['must'].append(range)

        # setup final wrapping timestamp aggregation
        timestamp_field = rule.get('timestamp_field', '@timestamp')
        timestamp_field = digest.get('timestamp_field', timestamp_field)

        agg_type = "date_histogram"
        timestamp_spec = self.get_aggregate_items(agg_type, [timestamp_field])
        timestamp_agg = timestamp_spec[0]

        # set interval (unless set)
        timestamp_interval = digest.get('timestamp_interval', '1h')
        key = list(timestamp_agg.keys())[0]

        log.debug("key=%s, agg_type=%s, timestamp_agg=%s", key, agg_type, timestamp_agg)
        if 'interval' not in timestamp_agg[key][agg_type]:
            timestamp_agg[key][agg_type]['interval'] = timestamp_interval

        first_value(timestamp_agg)['aggs'] = aggs

        result = {
            'query': query,
            'aggs': timestamp_agg,
            'size': 0
        }

        log.debug("result: %s", result)
        return result


    def get_aggregated_doc(self, bucket, doc=None):
        if doc is None:
            doc = {}

        for key, value in bucket.items():
            if not isinstance(value, dict): continue

            if 'value' in value:
                doc[key] = value['value']
            elif 'buckets' in value:
                for b in value['buckets']:
                    if 'key_as_string' in b:
                        doc[key] = b['key_as_string']
                    else:
                        doc[key] = b['key']

                    if '_id' not in doc:
                        doc['_id'] = b['key']

                    self.get_aggregated_doc(b, doc)
            else:
                doc[key] = value

        return doc


    def get_run_at(self):
        run_at = self.config.get('arguments.run_at')

        if not run_at:
            run_at = datetime.datetime.now()
        else:
            run_at = to_dt(run_at)

        return run_at

        return datetime.combine(run_at.date(), datetime.min.time())


    def get_relevant_indexes(self, rule, min_date, max_date):
        # get index exclude pattern
        index_exclude = rule.get('index.exclude', [])
        if isinstance(index_exclude, string):
            index_exclude = [index_exclude]

        index_exclude = [ re.compile(n.replace('*', '.*')) for n in index_exclude ]

        # predicate to check if indexname shall be excluded
        def is_excluded(index_name):
            for x in index_exclude:
                if x.match(index_name):
                    return True
            return False

        # setup final wrapping timestamp aggregation
        timestamp_field = rule.get('timestamp_field', '@timestamp')

        # get index pattern
        index_pattern = self.get_required(rule, 'index.source')

        # iterate over indexes and create a list of indexes to process
        index_names = {}
        for index_name in self.es.indices.get(index_pattern):
            if is_excluded(index_name): continue

            first_entry = self.es.search(index=index_pattern, query={
                "query": { "match_all": {} },
                "sort": [ { timestamp_field: { "order": "asc" } } ],
                "size": 1,
            })

            if not first_entry['total']: # nothing to do
                # index can be dropped
                continue

            first_entry = first_entry['hits']['hits'][0]['_source']
            startdate = first_entry[timestamp_field]

            if max_date:
                if startdate > max_date:
                    # this index must be skipped
                    continue

            latest_entry = self.es.search(index=index_name, query={
                 "query": { "match_all": {} },
                 "sort": [ { timestamp_field: { "order": "desc" } } ],
                 "size": 1,
            })

            last_entry = last_entry['hits']['hits'][0]['_source']
            enddate = to_dt(last_entry[timestamp_field])

            if enddate < min_date:
                # this index must be skipped
                continue

            index_names[index_name] = (startdate, enddate)

        return index_names



    def process(rule, action=None):
        # get minimum age of documents, default 14 days
        min_age = timedelta(*rule.get('age.min', {'days': 7}))
        max_date = end_of_day(self.get_run_day() - min_age)

        # maximum age of documents
        max_age = rule.get('age.max')
        if max_age:
            max_age = timedelta(*max_age)
            min_date = start_of_day(self.get_run_day() - max_age)

        # field specifying timestamp
        timestamp_field = rule.get('timestamp_field', '@timestamp')

        index_names = self.get_relevant_indexes(rule, min_date, max_date)

        # iterate from starttime to endtime in steps of timeframe
        timeframe = timedelta(self.config.get('digest.timeframe', {'days': 1}))

        _starttime = starttime
        while _starttime < endtime:
            # setup _endtime of current timeframe
            _endtime = _starttime + timeframe

            if _endtime > endtime:
                _endtime = endtime

            # collection to store failed index requests of aggregated docs
            failed = []

            # create list of indexes having documents within current timeframe
            indexes=[]
            for idx,idx_range in index_names.items():
                if idx_range[1] < _starttime or idx_range[0] > _endtime:
                    continue
                indexes.append(idx)

            # collection to store digestion_queries, important for later
            # handling of successful digestions
            digestion_queries = []

            # digest all documents
            for digest in rule.get('digests', []):
                query = self.make_query(rule, digest, starttime, _endtime)
                result = self.es.search(index=",".join(indexes), query=query)

                for agg in result['aggregations'][timestamp_field]['buckets']:
                    doc = self.get_aggregated_doc(agg['aggregations'])
                    target_index = timestamp.strftime(to_dt(doc[timestamp_field]))
                    doc['_index'] = target_index
                    bulk_success, bulk_failed = self.bulk_append(doc)
                    failed += bulk_failed

                digestion_queries.append(query['query'])

                if digest.get('on_success'):
                    bulk_success, bulk_failed = self.bulk_flush()
                    failed += bulk_failed

                    self.on_success(failed, digest, indexes, query['query'])

            bulk_success, bulk_failed = self.bulk_flush()
            failed += bulk_failed

            success_query = { 'bool': { 'should': digestion_queries } }
            self.on_success(failed, rule, indexes, success_query)

            delete_empty_indices = self.config.get('delete_empty_indices')
            if rule.get('delete_empty_indices', delete_empty_indices):
                self.delete_empty_indices(indexes)

            _starttime = _endtime


    def delete_documents(self, indexes, query):
        result = self.es.delete_by_query(
            index=",".join(indexes),
            body={'query': query}
        )
        log.info("deleted %s documents out of %s matching", result['deleted'], result['total'])


    def on_success(self, failed, rule, indexes, query):
        if len(failed):
            log.warning("there were erros doing bulk indexing: %s", failed)
            return

        for action in rule.get('on_success', []):
            if action in ('delete_digested', 'delete'):
                self.delete_documents(indexes, query)
            #body={'query': {'bool': {
            #    'should': digestion_queries
            #}}}

            elif isinstance(action, dict):
                assert 'delete' in action, "you must specify 'delete' item"

                range = self.make_range_query(_starttime, _endtime)
                if action['delete'] in ('all', '_all', '*'):
                    self.delete_documents(indexes, range)

                elif isinstance(action['delete'], (list,dict)):
                    _delete = action['delete']
                    if isinstance(_delete, dict):
                        _delete = [_delete]

                    self.delete_documents(indexes,
                        {'bool': {'must': _delete + [range]}})

                else:
                    raise NotImplemented(
                        "You cannot specify a deletion with %s" %
                        action['delete'])


    def delete_empty_indices(self, indexes):
        for idx in indexes:
            result = self.es.search(index=idx, query={
                 "query": { "match_all": {} },
                 "size": 0,
            })
            if result['hits']['total'] > 0:
                continue

            log.info("delete empty index %s", idx)

            self.indices.delete(idx)


    @classmethod
    def run_query(cls, config):
        config = ConfigFactory(config['config']).create(config)
        digester = Digester(config)

        digester.query()



#        for digest in
#        self.make_query()


#    def process_rules(self):








#!/usr/bin/env python
from distutils.core import setup

setup(
    name         = 'elastico',
    version      = '0.0.70',
    description  = "Elasticsearch Companion - a commandline tool",
    author       = "Kay-Uwe (Kiwi) Lorenz",
    author_email = "kiwi@franka.dyndns.org",
    url          = 'https://github.com/klorenz/python-elastico',
    license      = "MIT",

    install_requires=[
        'elasticsearch',
        'PyYAML',
        'pyaml',
        'requests',
        'argdeco',
        'markdown',
        'pytz',
        'python-dateutil',
    ],
    packages=[
        'elastico', 'elastico.cli'
    ],

    # package_data = {
    #    elastico: ['subfolder/*.x', ...]
    # }
    # include_package_data = True

    entry_points={
        'console_scripts': [
            'elastico = elastico.cli:main',
        ]
    }
)
